@@ -38,7 +38,7 @@ proc Doc_Root {{real {}}} {
 # Add a file system to the virtual document hierarchy
 
 proc Doc_AddRoot {virtual directory} {
-    Url_PrefixInstall $virtual [list DocDomain $directory]
+    Url_PrefixInstall $virtual [list DocDomain $virtual $directory]
 }
 
 # Define the index file for a directory
@@ -163,7 +163,7 @@ proc Doc_File {sock curfile npath} {
 # to return the contents.  The call to DocHandle is cached
 # by Url_Handle for future accesses to this url.
 
-proc DocDomain {directory sock suffix} {
+proc DocDomain {virtual directory sock suffix} {
     global Doc
     upvar #0 Httpd$sock data
 
@@ -204,7 +204,7 @@ proc DocDomain {directory sock suffix} {
     if {![DocFallback $path $suffix $cookie $sock]} {
 	# Couldn't find anything.
 	# check for cgi script in the middle of the path
-	Cgi_Domain $directory $sock $suffix
+	Cgi_Domain $virtual $directory $sock $suffix
     }
 }
 
@@ -219,7 +219,12 @@ proc DocFallback {path suffix cookie sock} {
     }
     set ok {}
     set accept [DocAccept $sock]
+
     foreach choice [glob -nocomplain $root.*] {
+	if {[string compare [file root $choice] $root] != 0} {
+	    # This prevents "foo.html.old" from matching for "foo.html"
+	    continue
+	}
 	set type [Mtype $choice]
 	if {[DocMatch $accept $type] && ![DocExclude $choice]} {
 	    lappend ok $choice
@@ -350,9 +355,17 @@ proc DocHandle {path suffix cookie sock} {
 
 proc DocDirectory {path suffix cookie sock} {
     upvar #0 Httpd$sock data
-    global Doc
+    global Doc tcl_platform
+
+    # Special case because glob doesn't work in wrapped files
+    # Just set indexpat to "index.tml" or "index.html"
+
     set npath [file join $path $Doc(indexpat)]
-    set newest [DocLatest [glob -nocomplain $npath]]
+    if {[info exist tcl_platform(isWrapped)] && $tcl_platform(isWrapped)} {
+	set newest $npath
+    } else {
+	set newest [DocLatest [glob -nocomplain $npath]]
+    }
     if {[string length $newest]} {
 	if {[string compare $Doc(tmlSuffix) [file extension $newest]] == 0} {
 	    foreach try [list [file root $newest].html [file root $newest].htm] {
@@ -371,12 +384,11 @@ proc DocDirectory {path suffix cookie sock} {
 }
 
 proc DocLatest {files} {
-    set mtime 0
     set newest {}
     foreach file $files {
 	if {[file readable $file]} {
 	    set m [file mtime $file]
-	    if {$m > $mtime} {
+	    if {![info exist mtime] || ($m > $mtime)} {
 		set mtime $m
 		set newest $file
 	    }
@@ -541,24 +553,20 @@ proc DocTemplate {sock template htmlfile suffix dynamicVar {interp {}}} {
     # Populate the global "page" array with state about this page
 
     set root ""
-    set prefix ""
     foreach d $dirs {
 	append root ../
-	append prefix /$d
     }
     if {[string length $htmlfile]} {
-	set url $prefix/[file tail $htmlfile]
 	set filename $htmlfile
 	set dynamic 0
     } else {
-	set url $prefix/[file tail $template]
 	set filename $template
 	set dynamic 1
     }
     interp eval $interp {uplevel #0 {catch {unset page}}}
 
     interp eval $interp [list uplevel #0 [list array set page [list \
-	url		$url 		\
+	url		$data(url)	\
 	template 	$template	\
 	filename	$filename	\
 	root		$root		\
@@ -581,8 +589,14 @@ proc DocTemplate {sock template htmlfile suffix dynamicVar {interp {}}} {
 
     # Check query data
     # steve: 5/8/98: Add multipart document upload handling
+    # If there is out-standing (i.e., unread) post data, then
+    # calling Url_DecodeQuery will automatcally read it
+    # and append to data(query)
 
-    if {[info exists data(query)] && [string length $data(query)]} {
+    if {[Httpd_PostDataSize $sock] > 0 && ![info exists data(query)]} {
+	set data(query) {}
+    }
+    if {[info exist data(query)]} {
 	set queryType application/x-www-urlencoded
 	set qualifiers {}
 	catch {
@@ -609,7 +623,7 @@ proc DocTemplate {sock template htmlfile suffix dynamicVar {interp {}}} {
 
     # Process the template itself
 
-    set html [DocSubst $template $interp]
+    set code [catch {DocSubst $template $interp} html]
 
     # Save return cookies, if any
 
@@ -619,6 +633,12 @@ proc DocTemplate {sock template htmlfile suffix dynamicVar {interp {}}} {
 	foreach c $cookie {
 	    Httpd_SetCookie $sock $c
 	}
+	interp eval $interp {uplevel #0 {unset page(set-cookie)}}
+    }
+
+    if {$code != 0} {
+	global errorCode errorInfo
+	return -code $code -errorcode $errorCode -errorinfo $errorInfo
     }
 
     set dynamic [interp eval $interp {uplevel #0 {set page(dynamic)}}]
@@ -649,7 +669,9 @@ proc Doc_Dynamic {} {
 
 # Doc_Cookie
 #
-#	Return a cookie value, if present, else ""
+#	Return a *list* of cookie values, if present, else ""
+#	It is possible for multiple cookies with the same key
+#	to be present, so we return a list.
 #
 # Arguments:
 #	cookie	The name of the cookie (the key)
@@ -657,15 +679,16 @@ proc Doc_Dynamic {} {
 
 proc Doc_Cookie {cookie} {
     global env
+    set result ""
     if {[info exist env(HTTP_COOKIE)]} {
 	foreach pair [split $env(HTTP_COOKIE) \;] {
 	    lassign [split [string trim $pair] =] key value
 	    if {[string compare $cookie $key] == 0} {
-		return $value
+		lappend result $value
 	    }
 	}
     }
-    return ""
+    return $result
 }
 
 # Doc_SetCookie
@@ -716,6 +739,18 @@ proc Doc_IsLinkToSelf {url} {
     return [expr {[string compare $url $page(url)] == 0}]
 }
 
+# Trigger a page redirect
+
+proc Doc_Redirect {newurl} {
+    return -code error \
+	    -errorcode  [list HTTPD_REDIRECT $newurl] \
+	    "Redirect to $newurl"
+}
+
+proc Doc_RedirectSelf {newurl} {
+    Doc_Redirect [Httpd_SelfUrl $newurl]
+}
+
 # Check modify times on all templates that affect a page
 
 proc DocCheckTemplate {template htmlfile} {
@@ -724,7 +759,7 @@ proc DocCheckTemplate {template htmlfile} {
     if {[file exists $htmlfile]} {
 	set mtime [file mtime $htmlfile]
     } else {
-	set mtime 0
+	return 1
     }
 
     # Look for .tml library files down the hierarchy.
